
// Schedules for deterministic communication

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "Procs.h"
#include "threads.h"
#include "adtags.h"
#include "schedule.h"
#include "MessSchedule.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `MessSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

MessSchedule :: MessSchedule(const int _elementLen) : elementLen(_elementLen) {
  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

  // g++ v2.7.2 won't accept this---internal compiler error?!
  //locIdxs = new IndexList [nProcs] ;
  //srvIdxs = new IndexList [nProcs] ;

  locIdxs = new IndexList [2 * nProcs] ;
  srvIdxs = locIdxs + nProcs ;
}

void MessSchedule :: build() {

  // Set `locList' and `srvList' to hop-ordered lists of non-empty messages,
  // and allocate contiguous buffers big enough to hold largest messages.

  int locBufSize = 0, srvBufSize = 0 ;
  for(int hop = 0 ; hop < nProcs ; hop++) {
    {
      const int remId = (locId + hop) % nProcs ;
      int size = locIdxs [remId].size() ;
      if(size) {
	locList.add(remId) ;
        if(size > locBufSize) locBufSize = size ;
      }
    }
    {
      const int remId = (locId - hop + nProcs) % nProcs ;
      int size = srvIdxs [remId].size() ;
      if(size) {
	srvList.add(remId) ;
        if(size > srvBufSize) srvBufSize = size ;
      }
    }
  }
  locBuf = new char [locBufSize * elementLen] ;  // *** potential word-
  srvBuf = new char [srvBufSize * elementLen] ;  //     alignment problem ***
}

MessSchedule :: ~MessSchedule() {
  delete [] srvBuf ;
  delete [] locBuf ;

  // See remark above re g++ v2.7.2
  //delete [] srvIdxs ;
  delete [] locIdxs ;
}

void MessSchedule :: gather(void* locBase, void* srvBase) {

  // Spawn server to send messages in `srvList'...

  SendServer sndServer(srvList.size(), srvList.data(),
                       srvBuf, srvIdxs, srvBase, elementLen) ;

  // Receive messages in `locList'...

  const int numMes = locList.size() ;
  int* remIds      = locList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = locIdxs + remId ;
    const int count = indexVec->size() ;
    int*    offsets = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(locBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    scatterOffBuf((char*) locBase, offsets, count, locBuf, elementLen) ;
  }

  // Wait for `sndServer' to terminate.

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: scatter(void* locBase, void* srvBase) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    scatterOffBuf((char*) srvBase, offsets, count, srvBuf, elementLen) ;
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: combine(void* locBase, void* srvBase,
                             void (*op)(void*, void*)) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    char* buf = srvBuf ;
    for(int i = 0 ; i < count ; i++) {
      (*op)((char*) srvBase + elementLen * offsets [i], buf) ;
      buf += elementLen ;
    }
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `SendServer' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

SendServer :: SendServer(const int _numMes, int* _remIds,
			 char* _sendBuf, IndexList* _sendIdxs,
			 void* _sendBase, int _elementLen) :
                     numMes(_numMes), remIds(_remIds),
                     sendBuf(_sendBuf), sendIdxs(_sendIdxs),
                     sendBase(_sendBase), elementLen(_elementLen) {
  mes = 0 ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

SendServer :: ~SendServer() {}

void SendServer :: transition(MPI_Status*) {
  mes++ ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

int SendServer :: active() const {
  return mes < numMes ;
}

void SendServer :: initSend(const int remId) {
  IndexList* indexVec = sendIdxs + remId ;
  const int count = indexVec->size() ;
  int*    offsets = indexVec->data() ;

  gatherOffBuf((char*) sendBase, offsets, count, sendBuf, elementLen) ;

  MPI_Isend(sendBuf, count * elementLen, MPI_CHAR,
	    remId, DAT, MPI_COMM_WORLD, event) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockMessSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockMessSchedule :: run(List<MessQueue>* dstMessages,
                              List<MessQueue>* srcMessages,
                              MultiBlockList* dstCopies,
                              MultiBlockList* srcCopies,
                              void* dstBase, void* srcBase) {

  // Do local copying...

  MultiBlock dstBlock = dstCopies->data() ;
  MultiBlock srcBlock = srcCopies->data() ;

  for(int i = 0 ; i < cpyNum ; i++) {
    copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen) ;
    nextBlock(dstBlock) ;
    nextBlock(srcBlock) ;
  }

  // Do sends and receives...

  //align_t sendBuf [(STATIC_BUFFER_SIZE + ALIGN_SZ - 1) / ALIGN_SZ] ;

  List<MessQueue>* rcvRecord = dstMessages ;
  List<MessQueue>* sndRecord = srcMessages ;

  int rcvRemId, rcvHop, rcvNumMes, rcvIMes ;
  int sndRemId, sndHop, sndNumMes, sndIMes ;
  Mess* rcvMessage ;
  Mess* sndMessage ;
  MultiBlock rcvBlock, sndBlock ;

  if(rcvRecord) {
    rcvRemId   = rcvRecord->remId ;
    rcvHop     = rcvRecord->hop ;
    rcvNumMes  = rcvRecord->messages.size() ;
    rcvMessage = rcvRecord->messages.data() ;
    rcvBlock   = rcvRecord->blocks.data() ;

    rcvIMes = 0 ;
  }
  else
    rcvMessage = 0 ;

  if(sndRecord) {
    sndRemId   = sndRecord->remId ;
    sndHop     = sndRecord->hop ;
    sndNumMes  = sndRecord->messages.size() ;
    sndMessage = sndRecord->messages.data() ;
    sndBlock   = sndRecord->blocks.data() ;

    sndIMes = 0 ;
  }
  else
    sndMessage = 0 ;

  MPI_Status status ;

  while(rcvMessage || sndMessage) {
    if(!sndMessage || (rcvMessage && rcvHop < sndHop)) {

      // There is no next send, or it has lower priority than the next receive.

      const int rcvNumBlk = rcvMessage->num  ;
      const int rcvSize   = rcvMessage->size ;

      if(rcvNumBlk == 1 && rcvBlock [MBLK_CTG]) {
	char* rcvBlockBase = (char*) dstBase +
                              elementLen * rcvBlock [MBLK_OFF] ;

	MPI_Recv(rcvBlockBase, rcvSize, MPI_CHAR, rcvRemId, DAT,
                 MPI_COMM_WORLD, &status) ;

	nextBlock(rcvBlock) ;
      }
      else {
	MPI_Recv(statBuf, rcvSize, MPI_CHAR, rcvRemId, DAT,
                 MPI_COMM_WORLD, &status) ;

	char* rcvPtr = (char*) statBuf ;
	for(int j = 0 ; j < rcvNumBlk ; j++) {
	  scatterMblkBuf(dstBase, rcvBlock, rcvPtr, elementLen) ;

	  rcvPtr += rcvBlock [MBLK_SIZ] ;
	  nextBlock(rcvBlock) ;
	}
      }

      rcvMessage++ ;
      rcvIMes++ ; 
      if(rcvIMes == rcvNumMes) {
	rcvRecord = rcvRecord->next ;
	if(rcvRecord) {
	  rcvRemId   = rcvRecord->remId ;
          rcvHop     = rcvRecord->hop ;
	  rcvNumMes  = rcvRecord->messages.size() ;
	  rcvMessage = rcvRecord->messages.data() ;
	  rcvBlock   = rcvRecord->blocks.data() ;

	  rcvIMes = 0 ;
	}
	else
	  rcvMessage = 0 ;
      }
    }
    else if(!rcvMessage || sndHop < rcvHop) {

      // There is no next receive, or it has lower priority than the next send.

      const int sndNumBlk = sndMessage->num  ;
      const int sndSize   = sndMessage->size ;

      if(sndNumBlk == 1 && sndBlock [MBLK_CTG]) {
	char* sndBlockBase = (char*) srcBase +
                              elementLen * sndBlock [MBLK_OFF] ;

	MPI_Send(sndBlockBase, sndSize, MPI_CHAR, sndRemId, DAT,
                 MPI_COMM_WORLD) ;

	nextBlock(sndBlock) ;
      }
      else {
	char* sndPtr = (char*) sendBuf ;
	for(int j = 0 ; j < sndNumBlk ; j++) {
	  gatherMblkBuf(srcBase, sndBlock, sndPtr, elementLen) ;

	  sndPtr += sndBlock [MBLK_SIZ] ;
	  nextBlock(sndBlock) ;
	}

	MPI_Send(sendBuf, sndSize, MPI_CHAR, sndRemId, DAT, MPI_COMM_WORLD) ;
      }

      sndMessage++ ;
      sndIMes++ ; 
      if(sndIMes == sndNumMes) {
	sndRecord = sndRecord->next ;
	if(sndRecord) {
	  sndRemId   = sndRecord->remId ;
          sndHop     = sndRecord->hop ;
	  sndNumMes  = sndRecord->messages.size() ;
	  sndMessage = sndRecord->messages.data() ;
	  sndBlock   = sndRecord->blocks.data() ;

	  sndIMes = 0 ;
	}
	else
	  sndMessage = 0 ;
      }
    }
    else {

      // Next send and next receive have the same priority.

      const int rcvNumBlk = rcvMessage->num  ;
      const int rcvSize   = rcvMessage->size ;

      const int sndNumBlk = sndMessage->num  ;
      const int sndSize   = sndMessage->size ;

      if(rcvNumBlk == 1 && rcvBlock [MBLK_CTG]) {
        if(sndNumBlk == 1 && sndBlock [MBLK_CTG]) {
	  char* rcvBlockBase = (char*) dstBase +
				elementLen * rcvBlock [MBLK_OFF] ;
	  char* sndBlockBase = (char*) srcBase +
				elementLen * sndBlock [MBLK_OFF] ;

	  MPI_Sendrecv(sndBlockBase, sndSize, MPI_CHAR, sndRemId, DAT,
                       rcvBlockBase, rcvSize, MPI_CHAR, rcvRemId, DAT,
                       MPI_COMM_WORLD, &status) ;

	  nextBlock(rcvBlock) ;
	  nextBlock(sndBlock) ;
        }
        else {
	  char* rcvBlockBase = (char*) dstBase +
				elementLen * rcvBlock [MBLK_OFF] ;

	  char* sndPtr = (char*) sendBuf ;
	  for(int j = 0 ; j < sndNumBlk ; j++) {
	    gatherMblkBuf(srcBase, sndBlock, sndPtr, elementLen) ;

	    sndPtr += sndBlock [MBLK_SIZ] ;
	    nextBlock(sndBlock) ;
	  }

	  MPI_Sendrecv(sendBuf, sndSize, MPI_CHAR, sndRemId, DAT,
                       rcvBlockBase, rcvSize, MPI_CHAR, rcvRemId, DAT,
                       MPI_COMM_WORLD, &status) ;

	  nextBlock(rcvBlock) ;
        }
      }
      else {
        if(sndNumBlk == 1 && sndBlock [MBLK_CTG]) {
	  char* sndBlockBase = (char*) srcBase +
				elementLen * sndBlock [MBLK_OFF] ;

	  MPI_Sendrecv(sndBlockBase, sndSize, MPI_CHAR, sndRemId, DAT,
                       statBuf, rcvSize, MPI_CHAR, rcvRemId, DAT,
                       MPI_COMM_WORLD, &status) ;

	  char* rcvPtr = (char*) statBuf ;
	  for(int j = 0 ; j < rcvNumBlk ; j++) {
	    scatterMblkBuf(dstBase, rcvBlock, rcvPtr, elementLen) ;

	    rcvPtr += rcvBlock [MBLK_SIZ] ;
	    nextBlock(rcvBlock) ;
	  }
	  nextBlock(sndBlock) ;
        }
        else {
          int j ;

	  char* sndPtr = (char*) sendBuf ;
	  for(j = 0 ; j < sndNumBlk ; j++) {
	    gatherMblkBuf(srcBase, sndBlock, sndPtr, elementLen) ;

	    sndPtr += sndBlock [MBLK_SIZ] ;
	    nextBlock(sndBlock) ;
	  }

	  MPI_Sendrecv(sendBuf, sndSize, MPI_CHAR, sndRemId, DAT,
                       statBuf, rcvSize, MPI_CHAR, rcvRemId, DAT,
                       MPI_COMM_WORLD, &status) ;

	  char* rcvPtr = (char*) statBuf ;
	  for(j = 0 ; j < rcvNumBlk ; j++) {
	    scatterMblkBuf(dstBase, rcvBlock, rcvPtr, elementLen) ;

	    rcvPtr += rcvBlock [MBLK_SIZ] ;
	    nextBlock(rcvBlock) ;
	  }
        }
      }

      rcvMessage++ ;
      rcvIMes++ ; 
      if(rcvIMes == rcvNumMes) {
	rcvRecord = rcvRecord->next ;
	if(rcvRecord) {
	  rcvRemId   = rcvRecord->remId ;
          rcvHop     = rcvRecord->hop ;
	  rcvNumMes  = rcvRecord->messages.size() ;
	  rcvMessage = rcvRecord->messages.data() ;
	  rcvBlock   = rcvRecord->blocks.data() ;

	  rcvIMes = 0 ;
	}
	else
	  rcvMessage = 0 ;
      }

      sndMessage++ ;
      sndIMes++ ; 
      if(sndIMes == sndNumMes) {
	sndRecord = sndRecord->next ;
	if(sndRecord) {
	  sndRemId   = sndRecord->remId ;
          sndHop     = sndRecord->hop ;
	  sndNumMes  = sndRecord->messages.size() ;
	  sndMessage = sndRecord->messages.data() ;
	  sndBlock   = sndRecord->blocks.data() ;

	  sndIMes = 0 ;
	}
	else
	  sndMessage = 0 ;
      }
    }
  }
}

//void BlockMessSchedule :: run(List<MessQueue>* dstMessages,
//                              List<MessQueue>* srcMessages,
//                              MultiBlockList* dstCopies,
//                              MultiBlockList* srcCopies,
//                              void* dstBase, void* srcBase) {
//
//  // Spawn server to send messages in `srcMessages'...
//
//  BlockSendServer sndServer(srcMessages, srcBase, elementLen) ;
//
//  // Do local copying...
//
//  MultiBlock dstBlock = dstCopies->data() ;
//  MultiBlock srcBlock = srcCopies->data() ;
//
//  for(int i = 0 ; i < cpyNum ; i++) {
//    copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen) ;
//    nextBlock(dstBlock) ;
//    nextBlock(srcBlock) ;
//  }
//
//  // Receive messages in `dstList'...
//
//  List<MessQueue>* record = dstMessages ;
//
//  while(record) {
//    const int  remId = record->remId ;
//    const int numMes = record->messages.size() ;
//
//    Mess*    message = record->messages.data() ;
//
//    MultiBlock block = record->blocks.data() ;
//
//    MPI_Status status ;
//
//    for(int mes = 0 ; mes < numMes ; mes++) {
//
//      const int num  = message->num  ;
//      const int size = message->size ;
//
//      if(num == 1 && block [MBLK_CTG]) {
//        char* blockBase  = (char*) dstBase + elementLen * block [MBLK_OFF] ;
//
//	MPI_Irecv(blockBase, size, MPI_CHAR,
//                  remId, DAT, MPI_COMM_WORLD, event) ;
//	Server :: waitClientEvent(&status) ;
//
//        nextBlock(block) ;
//      }
//      else {
//	MPI_Irecv(statBuf, size, MPI_CHAR,
//                  remId, DAT, MPI_COMM_WORLD, event) ;
//	Server :: waitClientEvent(&status) ;
//
//        char* ptr = (char*) statBuf ;
//
//        for(int j = 0 ; j < num ; j++) {
//          scatterMblkBuf(dstBase, block, ptr, elementLen) ;
//
//          ptr += block [MBLK_SIZ] ;
//          nextBlock(block) ;
//        }
//      }
//
//      message++ ;
//    }
//    
//    record = record->next ;
//  }
//
//  // Wait for `sndServer' to terminate.
//
//  while(sndServer.active())
//    Server :: waitServerTransition() ;
//}

void BlockMessSchedule :: show() {

  char buf [1024] ;

  // Local copying...

  MultiBlock dstBlock = locCopies.data() ;
  MultiBlock srcBlock = srvCopies.data() ;

  for(int i = 0 ; i < cpyNum ; i++) {
    //const int rank      = dstBlock [MBLK_RNK] ;
    int* dstStrs        = dstBlock + MBLK_STR ;
    int* srcStrs        = srcBlock + MBLK_STR ;
    int* exts           = dstStrs  + rank ;

    char* ptr = buf ;

    ptr += sprintf(ptr, "Copy (") ;
    if(rank) {
      ptr += sprintf(ptr, "%d", exts [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", exts [j]) ;
    }
    ptr += sprintf(ptr, ") * %d from %d + (", elementLen, srcBlock [MBLK_OFF]) ;
    if(rank) {
      ptr += sprintf(ptr, "%d", srcStrs [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", srcStrs [j]) ;
    }
    ptr += sprintf(ptr, ") to %d + (", dstBlock [MBLK_OFF]) ;
    if(rank) {
      ptr += sprintf(ptr, "%d", dstStrs [0]) ;
      for(int j = 1 ; j < rank ; j++)
        ptr += sprintf(ptr, ",%d", dstStrs [j]) ;
    }
    ptr += sprintf(ptr, ")") ;
   
    AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

    nextBlock(dstBlock) ;
    nextBlock(srcBlock) ;
  }

  List<MessQueue>* record ;

  // Messages to be sent...

  record = srvList ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;

      for(int j = 0 ; j < num ; j++) {
	//const int rank   = block [MBLK_RNK] ;
	int* strs        = block + MBLK_STR ;
	int* exts        = strs  + rank ;

	char* ptr = buf ;

	ptr += sprintf(ptr, "Send (") ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", exts [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", exts [k]) ;
	}
	ptr += sprintf(ptr, ") * %d from %d + (",
               elementLen, block [MBLK_OFF]) ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", strs [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", strs [k]) ;
	}
	ptr += sprintf(ptr, ") to proc %d", remId) ;
       
	AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

	nextBlock(block) ;
      }

      message++ ;
    }
    
    record = record->next ;
  }

  // Messages to be recieved...

  record = locList ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;

      for(int j = 0 ; j < num ; j++) {
	//const int rank   = block [MBLK_RNK] ;
	int* strs        = block + MBLK_STR ;
	int* exts        = strs  + rank ;

	char* ptr = buf ;

	ptr += sprintf(ptr, "Recv (") ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", exts [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", exts [k]) ;
	}
	ptr += sprintf(ptr, ") * %d to %d + (", elementLen, block [MBLK_OFF]) ;
	if(rank) {
	  ptr += sprintf(ptr, "%d", strs [0]) ;
	  for(int k = 1 ; k < rank ; k++)
	    ptr += sprintf(ptr, ",%d", strs [k]) ;
	}
	ptr += sprintf(ptr, ") from proc %d", remId) ;
       
	AdlibPrintf("BlockMessSchedule :: show: %s\n", buf) ;

	nextBlock(block) ;
      }

      message++ ;
    }
    
    record = record->next ;
  }
}


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////    The `BlockSendServer' class.                                     //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//BlockSendServer :: BlockSendServer(List<MessQueue>* sendList,
//                                   void* _sendBase, int _elementLen) :
//		         record(sendList),
//			 sendBase(_sendBase),
//                         elementLen(_elementLen) {
//  if(record) {
//    remId   = record->remId ;
//    numMes  = record->messages.size() ;
//
//    mes     = 0 ;
//    message = record->messages.data() ;
//
//    block   = record->blocks.data() ;
//
//    initSend() ;
//  }
//}
//
//void BlockSendServer :: transition(MPI_Status*) {
//  mes++ ;
//  if(mes < numMes) {
//    message++ ;
//
//    initSend() ;
//  }
//  else {
//    record = record->next ;
//
//    if(record) {
//      remId   = record->remId ;
//      numMes  = record->messages.size() ;
//
//      mes     = 0 ;
//      message = record->messages.data() ;
//
//      block   = record->blocks.data() ;
//
//      initSend() ;
//    }
//  }
//}
//
//void* BlockSendServer :: active() const {
//  return record ;
//}
//
//void BlockSendServer :: initSend() {
//  const int num  = message->num  ;
//  const int size = message->size ;
//
//  if(num == 1 && block [MBLK_CTG]) {
//    char* blockBase  = (char*) sendBase + elementLen * block [MBLK_OFF] ;
//    nextBlock(block) ;
//
//    MPI_Isend(blockBase, size, MPI_CHAR,
//	      remId, DAT, MPI_COMM_WORLD, event) ;
//  }
//  else {
//    char* ptr = (char*) sendBuf ;
//
//    for(int j = 0 ; j < num ; j++) {
//      gatherMblkBuf(sendBase, block, ptr, elementLen) ;
//
//      ptr += block [MBLK_SIZ] ;
//      nextBlock(block) ;
//    }
//
//    MPI_Isend(sendBuf, size, MPI_CHAR, remId, DAT, MPI_COMM_WORLD, event) ;
//  }
//}


// ----------------------------------------------------------------------
// Things to do
//
//     Change `MessSchedule' to use statically declared buffer.
//     General reorganizations to look more like modern `BlockMessSchedule'.
//
//     Replace `scatterOffBuf' by a virtual function, and unify implementations
//     of `gather', `scatter', `combine'.
//
//     Could `run' be a global function (as it was in the bad old
//     days), or a member of some class other than `BlockDataSchedule'?
//
//     The current implementation of `run' is very ugly.  It was changed
//     because MPICH/P4 apparently has a bad implementation of
//     `MPI_Waitany', which was driving down performance by an order of
//     magnitude.  The interim implementation uses blocking sends and
//     receives.  One simple improvement would be to revert to non-blocking
//     sends and receives, but use them deterministically.  Ultimately
//     it would presumably be better to use a true thread package, if
//     there was one that was efficient and portable.

